﻿//-----[ AA3DMANAGER ]-----------------------------------------------------
/* MANAGER CLASS for all things 3D, overalys on Three.js for specific
purposes that are custom to this game engine.
*/
//-------------------------------------------------------------------------
"use strict";
//-------------------------------------------------------------------------
// sin/cos lookup tables to avoid overhead of math lookups
//-------------------------------------------------------------------------
var sinTable = [];
var cosTable = [];
//-------------------------------------------------------------------------


//-----[ AA3DMANAGER ]-----------------------------------------------------
class AA3DManager {

     
    //-----[ AA3D MANAGER: CONSTRUCTOR ]-----------------------------------
    constructor() {

        for (var s = 0; s < 360; s++) {
            sinTable[s] = Math.sin(s * Math.PI / 180.0000);
            cosTable[s] = Math.cos(s * Math.PI / 180.0000);
        }

        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!AA3D_MANAGER) {
            AA3D_MANAGER = this;
        } else {
            console.log("AAENGINE: ERROR: Attempt to create multiple 3D managers." +
                        "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return AA3D_MANAGER;
        }
    }
    //--------------------------------------------------------------------- 

    
    //-----[ AA3D MANAGER: INIT ]------------------------------------------ 
    //--------------------------------------------------------------------- 
    init() {
        if (!this._initialized) {
            //create renderer and attach to the engine container element
            this._renderer = new THREE.WebGLRenderer();
            this._container = document.getElementById('engine_background');
            this._renderer.domElement.style.position = 'absolute';
            this._container.appendChild(this._renderer.domElement);
            this._renderer.setSize(
                GAME_WINDOW.GAME_WINDOW_WIDTH,
                GAME_WINDOW.GAME_WINDOW_HEIGHT);

            //set the clear color to black, and create an empty scene
            this._renderer.setClearColor(0x000000, 1.0);
            this._scene = new THREE.Scene();
            this._scene.scale.set(GAME_WINDOW_GLOBAL_SCALE,
                                  GAME_WINDOW_GLOBAL_SCALE,
                                  GAME_WINDOW_GLOBAL_SCALE);

            //create a default light for testing - TODO: clean this up later
            var light = new THREE.PointLight(0xFFFFFF);
            light.position.set(0, 0, 800);
            light.decay = 0.0001;
            this._scene.add(light);

            //create the main orthographic game camera, at the 
            //scale that aaron wants it...
            if (GAME_WINDOW._PLAYFIELD_ORIENTATION == 'horizontal') {
                this._camera = new THREE.OrthographicCamera(
                    -GAME_WINDOW_UNSCALED_RESOLUTION / 2, //left
                     GAME_WINDOW_UNSCALED_RESOLUTION / 2, //right
                     GAME_WINDOW_UNSCALED_RESOLUTION /
                        GAME_WINDOW.VIEW_ASPECT / 2, //top
                    -GAME_WINDOW_UNSCALED_RESOLUTION /
                        GAME_WINDOW.VIEW_ASPECT / 2, //bottom
                     1,        //near
                     10000);   //far
            } else {
                this._camera = new THREE.OrthographicCamera(
                    -GAME_WINDOW_UNSCALED_RESOLUTION *
                        GAME_WINDOW.VIEW_ASPECT / 2, //left
                     GAME_WINDOW_UNSCALED_RESOLUTION *
                        GAME_WINDOW.VIEW_ASPECT / 2, //right
                     GAME_WINDOW_UNSCALED_RESOLUTION / 2, //top
                    -GAME_WINDOW_UNSCALED_RESOLUTION / 2, //bottom
                     1,        //near
                     10000);   //far
            }

            this._camera.position.x = 0;
            this._camera.position.y = 0;
            this._camera.position.z = 1000; //arbitrary given ortho

            this._camera.lookAt(this._scene.position);

            //if debugging, set vendor, version info into display
            if (DEBUG) {
                DEBUG_DISPLAY._vendorDisplay.nodeValue =
                    this._renderer.context.getParameter(
                    this._renderer.context.VENDOR);
                DEBUG_DISPLAY._versionDisplay.nodeValue =
                    this._renderer.context.getParameter(
                    this._renderer.context.VERSION);
                DEBUG_DISPLAY._rendererDisplay.nodeValue =
                    this._renderer.context.getParameter(
                    this._renderer.context.RENDERER);
                var s = this._renderer.context.getParameter(
                    this._renderer.context.SHADING_LANGUAGE_VERSION).toString();
                DEBUG_DISPLAY._sLversionDisplay.nodeValue =
                    s.slice(0, 33);
            }
           
            this._gameObjectsArray = []; //array of all game objects in scene
            this._initialized = true;

            this.render();
        } else {
            //this can be considered a 'reset' of the 3D manager, but
            //things already exist.  error for now...
            console.log("AAENGINE: Init called more than once on 3D Manager.");
        }
    }
    //--------------------------------------------------------------------- 


    //-----[ AA3D MANAGER: RESIZE ]---------------------------------------- 
    //--------------------------------------------------------------------- 
    resize() {

        if (this._renderer) {
            this._renderer.setSize(
                GAME_WINDOW.GAME_WINDOW_WIDTH,
                GAME_WINDOW.GAME_WINDOW_HEIGHT);

            this._renderer.setViewport(0, 0,
                GAME_WINDOW.GAME_WINDOW_WIDTH,
                GAME_WINDOW.GAME_WINDOW_HEIGHT);

            this._camera.left = GAME_WINDOW.GAME_WINDOW_WIDTH / -2;
            this._camera.right = GAME_WINDOW.GAME_WINDOW_WIDTH / 2;
            this._camera.top = GAME_WINDOW.GAME_WINDOW_HEIGHT / -2;
            this._camera.bottom = GAME_WINDOW.GAME_WINDOW_HEIGHT/2;

        }

        if (!USE_THREADING && this._gameObjectsArray) {
            for (var i = 0; i < this._gameObjectsArray.length; i++) {
                this._gameObjectsArray[i].resize();
            }
        }
    }
    //--------------------------------------------------------------------- 


    //-----[ AA3D MANAGER: RENDER ]----------------------------------------
    /* PURPOSE: Renders the scene.  Should be called after all manipulations
    of anything that uses game-object as a base, and generally 'last' per
    frame.  Will draw all objects in the scene.
    
    TODO: expand this for various subsystems like particles and whatnot
    */
    //--------------------------------------------------------------------- 
    render() {
        this._renderer.clear();
        this._renderer.render(this._scene, this._camera);
    }
    //--------------------------------------------------------------------- 

    updateAllGameObjects() {
        for (var i = 0, j = this._gameObjectsArray.length; i < j; i++) {

            //call the update function of all objects in the game objects
            //array.  
            this._gameObjectsArray[i].update();

            //NOTE: Render is always called last in the update function.
            //(you can toggle rendering based on internal class logic if you want,
            //but note that this means it simply won't update where the 3D system
            //draws it.  If you need to hide it, move it offscreen, block visibility,
            //or remove it from the world)
            this._gameObjectsArray[i].render();
        }
    }
    //-----[ AA3D MANAGER: ADDGAMEOBJ ]------------------------------------ 
    //PURPOSE: Adds a game object to the scene.  Game objects that change
    //their mesh should delete their old mesh from the scene and add their
    //new one (unless a pooling system for common meshes is used later)
    //--------------------------------------------------------------------- 
    addGameObject(gameObjType) {
        if (USE_THREADING) {
            THREAD_MANAGER.addGameObject(gameObjType);
        } else {
            var g = eval("new " + gameObjType + "();");
            var i = this._gameObjectsArray.push(g);
            g.setID(i - 1);
        }
    }
    //---------------------------------------------------------------------

    //-----[ AA3D MANAGER: GET RENDERER ]----------------------------------
    // returns a reference to the renderer 
    //---------------------------------------------------------------------
    get RENDERER() { return this._renderer; }
    get GAME_OBJECTS() { return this._gameObjectsArray; }
    get SCENE() { return this._scene; }
}
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// creation of global THREAD_MANAGER -- initialization called from
// main application init
//-------------------------------------------------------------------------
var AA3D_MANAGER = AA3D_MANAGER || new AA3DManager();
//-------------------------------------------------------------------------