﻿//-------[ GAME OBJECT ]---------------------------------------------------
/* Purpose: base game object class from which all others extend.  This is
the core class that relates game objects to screen positions, collision,
bins, etc.  There are 3 main types of game object under the hood:
0 = box (width and height are separate, but no rotation of bounding box)
1 = circle (a diameter is provided for bounding circle, and a w/h/l of the sphere)
2 = n-gon (an arbitrary convex polygon)

NOTE: a special case of an n-gon is one with only 2 points (ie a line) that
is commonly used for projectiles as the fastest approximate shape.
*/
//-------------------------------------------------------------------------
"use strict";


//-------------------------------------------------------------------------
class GameObject {
    //------[  GAME OBJECT::Constructor ]----------------------------------
    // Purpose: Create a game object.  Everything is in initialize routines
    //---------------------------------------------------------------------
    constructor() {
        this._typeID = 0;
        this._initialized = false;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::setID ]----------------------------------------
    // Purpose: set the ID of this game object.
    //---------------------------------------------------------------------
    setID(id) {
        this._id = id;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::setMeshID ]------------------------------------
    // Purpose: set the mesh ID of this game object.
    //---------------------------------------------------------------------
    setMeshID(id) {
        this._meshID = id;
    }
    //---------------------------------------------------------------------

    //------[  GAME OBJECT::setTypeID ]------------------------------------
    // Purpose: set the type ID of this game object for where it slots
    // in collision detection.
    // //0 = player, 1=playerBullet, 2=enemy, 3=enemyBullet
    //---------------------------------------------------------------------
    setTypeID(id) {
        this._typeID = id; 
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::Shared Init ]----------------------------------
    // Purpose: set base geometry and material, set default pos/rot/scale
    //---------------------------------------------------------------------
    sharedInit() {

        this._position = new THREE.Vector3(0, 0, 0);
        this._scale = new THREE.Vector3(1, 1, 1);
        this._rotation = new THREE.Vector3(0, 0, 0);
        this._meshScale = new THREE.Vector3(1, 1, 1);

        this._collidedThisFrame = false;
        this._collisions = [];
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::InitAsBox ]------------------------------------
    // Purpose: set w/h, type, and init flag as box
    // w = 2d width, h = 2dheight, sx = scaleX, sy = scaleY, sz = scaleZ
    //---------------------------------------------------------------------
    initAsBox(w, h, sx, sy, sz) {

        this.sharedInit(); //create base attribs

        this._2dWidth = w;  //unprojected width in 2D
        this._2dHeight = h; //unprojected height in 2D

        this._adjusted_points = [];
        for (var i = 0; i < 4; i++) {
            this._adjusted_points[i] = [];
        }
        //set flag that this object is initialized, as type box
        this._type = 0;//0 = box, 1 = circle/sphere, 2 = n-gon
        this._initialized = true;

        this.initialScaleWidth = sx;
        this.initialScaleHeight = sy;
        this.initialScaleDepth = sz;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::InitAsCircle ]---------------------------------
    // Purpose: set w/h, type, and init flag as circle
    //---------------------------------------------------------------------
    initAsCircle(r, sx, sy, sz) {

        this.sharedInit();
        this._radius = r;
        this._2dWidth = r;
        this._2dHeight = r;

        this.initialScaleWidth = sx;
        this.initialScaleHeight = sy;
        this.initialScaleDepth = sz;

        this._adjusted_points = [];
        for (var i = 0; i < 4; i++) {
            this._adjusted_points[i] = [];
        }
        //set flag that this object is initialized, as type box
        this._type = 1;//0 = box, 1 = circle/sphere, 2 = n-gon
        this._initialized = true;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::InitAsPoly ]-----------------------------------
    // Purpose: set w/h, type, and init flag as box
    //---------------------------------------------------------------------
    initAsPoly(w, h, sx, sy, sz, points) {

        this.sharedInit(); //create base attribs

        this._2dWidth = w; //unprojected width in 2D
        this._2dHeight = h; //unprojected height in 2D

        //list of 2D poly points
        this._points = points;
        this._adjusted_points = [];
        this._collision_points = [];
        for (var i = 0; i < this._points.length; i++) {
            this._adjusted_points[i] = [0, 0];
            this._collision_points[i] = [0, 0];
        }
        this._points_length = this._points.length;

        this.initialScaleWidth = sx;
        this.initialScaleHeight = sy;
        this.initialScaleDepth = sz;
        //set flag that this object is initialized
        this._type = 2; //0 = box, 1 = sphere, 2 = n-gon
        this._initialized = true;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::Update ]---------------------------------------
    update() {
        //nothing.  The engine expects that update will be overridden
        //by any class that extends GameObject.  
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::Render ]---------------------------------------
    // Purpose: if this game object is initialized, this function is
    // calculates the 2d and 3d position, rotation and scale based on the
    // position, rotation, and scale vectors that are manipulated by 
    // objects that extend this class.  This also updates 2D and 3D positions
    // based on window scale.  Called from the 'updateAllGameObjects' 
    // function in the 3D Manager of the game engine, or via a management
    // thread if threading is enabled.  Always called AFTER
    // the update function of a class that extends this base class.
    //---------------------------------------------------------------------
    render() {
        
        if (this._initialized) {

            //-------------------------------------------------------------
            //NOTE: 3D rendering actually happens in 3D manager, 
            //because the object is added to a scene and rendered from there.

            //reset collision data this frame (to be updated by collision
            //manager after updates are done)
            this._collidedThisFrame = false;
            this._collisions = [];
            

            //-------------------------------------------------------------
            // update 2d positions and scale effects
            //-------------------------------------------------------------
            this._2dPosX = this._position.x;
            this._2dPosY = this._position.y;
            
            this._tx = (this._position.z * GAME_WINDOW_ZOOM_EFFECT);
            this._ty = (this._position.z * GAME_WINDOW_ZOOM_EFFECT);

            if (GAME_WINDOW.WINDOW_WIDTH >= GAME_WINDOW.WINDOW_HEIGHT) {
                this._tx *= (this._2dWidth / this._2dHeight);
            } else {
                this._ty *= (this._2dHeight / this._2dWidth);
            }

            this._width =  this._scale.x * (this._2dWidth +  this._tx);
            this._height = this._scale.y * (this._2dHeight + this._ty);

            this._tx = (this._width / this._2dWidth);
            this._ty = (this._height / this._2dHeight);
            
            //-------------------------------------------------------------
            //set the actual position/rotation/scale in the 3D system
            //NOTE: Moved this to the model manager to be async from game object updates

            //-------------------------------------------------------------
            //set a scale based on z-position for a 'zoom' effect in ortho
            if (this._position.z < 0 && (Math.abs(this._position.z) >= this._2dWidth ||
                                         Math.abs(this._position.z) >= this._2dHeight)) {
                //this is a trap case for when an object is farther away than it is wide
                //or tall in projection.  This would normally produce a negative / inverted 
                //scale, which is indeterminite.  So we trap these cases and just make them
                //very tiny instead.
                
                //this._mesh.scale.set(0.00000001*this._scale.x, 0.00000001*this._scale.y, 1);
                this._meshScale.x = 0.00000001 * this._scale.x;
                this._meshScale.y = 0.00000001 * this._scale.y;

                this._width = 1 * this._scale.x; this._height = 1 * this._scale.y;
                this._tx = 0.01 * this._scale.x; this._ty = 0.01 * this._scale.y;
                
            } else {
                //this._mesh.scale.set(this._tx * this._2dWidth, this._ty * this._2dHeight, 1);
                this._meshScale.x = this._tx * this.initialScaleWidth; 
                this._meshScale.y = this._ty * this.initialScaleHeight;

                //fudge factor so things don't look completely flattened
                this._meshScale.z = this._ty * this.initialScaleDepth;
            }

            //-------------------------------------------------------------
            //update screen coordinates from the unprojected 2D coords
            this._projected2dPosX = this._2dPosX * GAME_WINDOW.GAME_WINDOW_SCALE +
                              GAME_WINDOW.GAME_WINDOW_HALF_WIDTH;

            this._projected2dPosY = -this._2dPosY * GAME_WINDOW.GAME_WINDOW_SCALE +
                              GAME_WINDOW.GAME_WINDOW_HALF_HEIGHT;;

            //-------------------------------------------------------------
            //IF BOX
            if (this._type == 0) {
                //draw a box as a test...

                this._hw = this._width *0.5;
                this._hh = this._height * 0.5;
                this._adjusted_points[0][0] = -this._hw;
                this._adjusted_points[0][1] = -this._hh;
                this._adjusted_points[1][0] = this._hw;
                this._adjusted_points[1][1] = -this._hh;
                this._adjusted_points[2][0] = this._hw;
                this._adjusted_points[2][1] = this._hh;
                this._adjusted_points[3][0] = -this._hw;
                this._adjusted_points[3][1] = this._hh;

                if (DEBUG && !USE_THREADING) {
                    if (DEBUG_DISPLAY._2dDebugVisibility) {
                        DEBUG_DISPLAY.drawDebugBox(
                            this._projected2dPosX - this._hw
                            * GAME_WINDOW.GAME_WINDOW_SCALE, //top
                            this._projected2dPosY - this._hh
                            * GAME_WINDOW.GAME_WINDOW_SCALE, //left
                            this._width * GAME_WINDOW.GAME_WINDOW_SCALE, //width
                            this._height * GAME_WINDOW.GAME_WINDOW_SCALE, //height
                            'rgba(255,255,255,0.0)', 'rgba(255,255,255,1.0)');
                            //fill color, stroke color
                    }
                }
            }
            
            //IF SPHERE
            else if (this._type == 1) {

                //set radius based on adjusted size
                if (this._width > this._height) { this._radius = this._width*0.5 }
                else { this._radius = this._height*0.5; }

                this._hw = this._width * 0.5;
                this._hh = this._height * 0.5;
                this._adjusted_points[0][0] = -this._hw;
                this._adjusted_points[0][1] = -this._hh;
                this._adjusted_points[1][0] = this._hw;
                this._adjusted_points[1][1] = -this._hh;
                this._adjusted_points[2][0] = this._hw;
                this._adjusted_points[2][1] = this._hh;
                this._adjusted_points[3][0] = -this._hw;
                this._adjusted_points[3][1] = this._hh;
                
                if (DEBUG && !USE_THREADING) {
                    if (DEBUG_DISPLAY._2dDebugVisibility) {
                        DEBUG_DISPLAY.drawDebugCircle(
                            this._projected2dPosX,
                            this._projected2dPosY,
                            this._radius * GAME_WINDOW.GAME_WINDOW_SCALE,
                            'rgba(255,255,255,0.0)', 'rgba(255,255,255,1.0)');
                        //fill color, stroke color) 
                    }
                }
            }
            //-------------------------------------------------------------
            //IF POLY
            else if (this._type == 2) {
               
               //need to ROTATE POINTS for collision armature
               for (var i = 0; i < this._points_length; i++) {

                   this._x = this._points[i][0] * this._tx;
                   this._y = this._points[i][1] * this._ty;

                   //var r = this._rotation.z;
                   this._r = Math.floor((-this._rotation.z / (Math.PI / 180)) % 360);
                   if (this._r < 0) { this._r += 360; }

                   //this._adjusted_points[i][0] = x * Math.cos(r) - y * Math.sin(r);
                   //this._adjusted_points[i][1] = x * Math.sin(r) + y * Math.cos(r);
                   this._adjusted_points[i][0] = this._x * cosTable[this._r]
                                               - this._y * sinTable[this._r];
                   this._adjusted_points[i][1] = this._x * sinTable[this._r]
                                               + this._y * cosTable[this._r];

               }
               if (DEBUG  && !USE_THREADING) {
                   if (DEBUG_DISPLAY._2dDebugVisibility) {
                       //draw a poly as a test...
                       DEBUG_DISPLAY.drawDebugPoly(
                        this._projected2dPosX, //top
                        this._projected2dPosY, //left
                        this._adjusted_points, //points after scale and rotation...
                        GAME_WINDOW.GAME_WINDOW_SCALE,
                        'rgba(255,255,255,0.0)', 'rgba(255,255,255,1.0)'); //fill color, stroke color
                   }
               }
            }

            if (USE_THREADING) {
                //the thread will compile all game objects 
                //and update them across the bridge
                //to the model manager, so do nothing.
            } else {
                //update the game object with the model 
                //manager so it is at the correct
                //pos/rot/scale/etc. for rendering 
                AAMODEL_MANAGER.updateModel(this._id,
                    this._position, this._rotation,
                    this._meshScale, this._meshID);
            }
            //add this shape to the collision manager for collision pass!
            AACOLLISION_MANAGER.addObjectToBins(this);
           
        } else {
            console.log("AAENGINE: ERROR: Render called on unitialized game object.")
        }
    }

    //------[  GAME OBJECT::setAtUnscaled2dPos ]---------------------------
    // Purpose: set object at specific 2d position in the arbitrary coords
    // of the window system established in game window.  This is often a
    // set scale such as 1920x1080 or similar, that is then scaled up or
    // down based on the actual pixel density of the monitor.  
    // NOTE: you *usually* want this for things inside the world.
    //---------------------------------------------------------------------
    setAtUnscaled2dPos(x, y) {
        this._position.x = x;
        this._position.y = y;
        this._2dPosX = x;
        this._2dPosY = y;
    }
    //---------------------------------------------------------------------

    //------[  GAME OBJECT::setAt3dPosition]-------------------------------
    // Purpose: set object at specific 3D position, and update 2D position
    // to match.
    //---------------------------------------------------------------------
    setAt3DPosition(x, y, z) {
        this._position.set(x, y, z);
        this._2dPosX = x;
        this._2dPosY = y;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::setAt2dScreenPosition]-------------------------
    // Purpose: set object at specific screen position in raw window coords
    // NOTE: you *usually* want this for overlays that float near the 'edge'
    // of the window
    //---------------------------------------------------------------------
    setAtScreen2dPosition(x, y) {

        x = (x - GAME_WINDOW.GAME_WINDOW_HALF_WIDTH) /
                 GAME_WINDOW.GAME_WINDOW_SCALE;
        y = (y - GAME_WINDOW.GAME_WINDOW_HALF_HEIGHT) /
                 GAME_WINDOW.GAME_WINDOW_SCALE;
        this._position.x = x;
        this._position.y = -y;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::rotateToZ ]------------------------------------
    // Purpose: set rotation to a specific angle
    //---------------------------------------------------------------------
    rotateToZ(ang) {
        this._rotation.z = ang * Math.PI / 180.0;
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::resize ]---------------------------------------
    // Purpose: this is a default base class handler for resize events, can 
    // be overwritten for anything that needs to adjust on resize
    // events
    //---------------------------------------------------------------------
    resize() {
        //nothing
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::collide ]--------------------------------------
    // Purpose: this is a default base class handler for collisions.  this
    // is called directly from the collision manager, and calls the
    // 'collision' callback that should be implemented for all classes that
    // extend gameObject.
    //---------------------------------------------------------------------
    collide(gO) {
        
        var q = gO._id;

        //if we hae already collided with this object, then return
        for (var z = 0; z < this._collisions.length; z++) {
            if (q == this._collisions[z]) {
                return;
            }
        }

        //store this collider in list of collisions
        this._collisions.push(q);
        this._collidedThisFrame = true;

        if (DEBUG && !USE_THREADING) {
            if (DEBUG_DISPLAY._2dDebugVisibility) {
                if (this._type == 0) {
                    DEBUG_DISPLAY.drawDebugBox(
                        this._projected2dPosX - this._hw
                        * GAME_WINDOW.GAME_WINDOW_SCALE, //top
                        this._projected2dPosY - this._hh
                        * GAME_WINDOW.GAME_WINDOW_SCALE, //left
                        this._width * GAME_WINDOW.GAME_WINDOW_SCALE, //width
                        this._height * GAME_WINDOW.GAME_WINDOW_SCALE, //height
                        'rgba(255,0,0,0.6)', 'rgba(255,0,0,1.0)');

                } else if (this._type == 1) {
                    //circle
                    DEBUG_DISPLAY.drawDebugCircle(
                        this._projected2dPosX,
                        this._projected2dPosY,
                        this._radius * GAME_WINDOW.GAME_WINDOW_SCALE,
                        'rgba(255,0,0,0.6)', 'rgba(255,0,0,1.0)');

                } else if (this._type == 2) {
                    DEBUG_DISPLAY.drawDebugPoly(
                        this._projected2dPosX, //top
                        this._projected2dPosY, //left
                        this._adjusted_points, //points after scale and rotation...
                        GAME_WINDOW.GAME_WINDOW_SCALE,
                        'rgba(255,0,0,0.6)', 'rgba(255,0,0,1.0)');
                }
            }
        }

        //call up to main object a collision happened...
        this.collision(gO);
        

        
    }
    //---------------------------------------------------------------------


    //------[  GAME OBJECT::Get Functions ]--------------------------------
    // Purpose: set base geometry and material, set default pos/rot/scale
    //---------------------------------------------------------------------
    get POS2D() { return [this._2dPosX, this._2dPosY] } //2d point/array
    get POS3D() { return this._position } //3d pos / THREE.vector
    //adjusted position on screen after projection/scale - 2dPoint/array
    get SCREENPOS2D() { return [this._projected2dPosX, this._projected2dPosY] }
    get ROTATION() { return this._rotation } //rotation / THREE.vector
    get SCALE() { return this._scale } //scale / THREE.vector
    //---------------------------------------------------------------------
}