﻿//-------[ INPUT MANAGER ]-------------------------------------------------
/*  
    PURPOSE: Class for an INPUT MANAGER object that contains appropriate
    methods for handling mouse interactions, keyboard interactions, and
    gamepad interactions at the base level (ie events and information are
    captured, and this information is requested elsewhere for game logic)

    //-----[ MOUSE INPUT NOTES ]-------------------------------------------
    NOTES: Other parts of the application should always check mosue coords
    in combination with the INPUT_MANAGER.MOUSEACTIVE flag, which is false
    in situations where the mosue is not over the game window (and thus 
    values are erratic across different browsers and windowing systems)

     //-----[ KEYBOARD INPUT NOTES ]---------------------------------------
    /* PURPOSE: Reacts to keyboard events and looks for any keys we care about
    via keyCode.  If Arrow key or W/A/S/D is pressed, or space, then update the
    gKeys array.  NOTES: currently suffers from the typical limitation of the
    '3 key problem' on non-low-level keyboard polling (ie is event based not a 
    true polling solution, javascript doesn't have a 'isKeyDown()' equivalent).
    Also note that interactions between the two control key sets can get a bit
    strange at times (ie pressing 'W' and 'up arrow' and then letting up on
    one will clear the otehr from the buffer.  

    stored internally in an array [0-4] in the _key_status member var
    [W/up(bool), S/down(bool), A/left(bool), D/right(bool), SPACE(bool)]
    accessed through retrieval methods INPUT_MANAGER.KB_DOWN, .KB_UP, 
    .KB_LEFT, .KB_RIGHT, and .KB_SPACE

    //-----[ GAMEPAD INPUT NOTES ]-----------------------------------------

*/
//-------------------------------------------------------------------------
"use strict";
var CONTROLLER_DEAD_ZONE = 0.2; //deadzone for stick controls
var TRIGGER_DEAD_ZONE = 0.1;    //threshold for trigger actions

class InputManager {
    constructor() {
        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!INPUT_MANAGER) {
            INPUT_MANAGER = this;

            //block context menu on right click 
            //(so we can use right mosue button)
            window.oncontextmenu = function () { return false; }

            //initialization of variables for mouse handling
            this._lastGameMouseX =  -10000;  this._lastGameMouseY = -10000;
            this._rawMouseX =       -10000;  this._rawMouseY =      -10000;
            this._gameMouseX =      -10000;  this._gameMouseY =     -10000;
            this._screenMouseX =    -10000;  this._screenMouseY =   -10000;
            this._mouseActive = -1;
            this._mouseState = [0, 0, 0, 0];

            //array of key statuses and their status
            //0=W/UP, 1=S/DOWN, 2=D/RIGHT, 3=L/LEFT, 4=SPACE, 
            //this._key_status = [0, 0, 0, 0, 0];
            this._key_status = new Int16Array(5);
            this._text_controls = [0, 0, 0, 0, 0, 0]; //F / Q / P / C / Enter / Tab

            //array of gamepad devices (may or may not be connected)
            this._gamepads = [0, 0, 0, 0, 0, 0, 0, 0];

            //handles to bound game pads as player inputs (when active)
            this._gamepad_one = 0;
            this._gamepad_two = 0;

            //array of state variables for gamepad data (polled per frame)
            this._gamepad_one_data = new Float32Array(23);
                //[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
            this._gamepad_two_data = new Float32Array(23);
                //[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
            this._mouse_data = new Int32Array(3);

            for (var i = 0; i < 23; i++) {
                this._gamepad_one_data[i] = 0;
                this._gamepad_two_data[i] = 0;
            }

            //detect controllers that are already connected at startup
            this.connectController();

            //storage of 'how much' the mosue wheel has been scrolled
            this._wheelDeltaStore = 0;
            this._mouseWheelSensitivity = 3; //higher is less sensitive
            this._totalWheelDeltaStore = 0;
            this._wheelClearDeltaStore = 0;

            this._sendData = false;

        } else {
            console.log(
            "AAENGINE: ERROR: Attempt to create multiple input managers." +
            "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return INPUT_MANAGER;
        }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::INIT ]---------------------------------------
    // initialization routine for anything that cannot be pre-defined 
    // at load time (ie that relies on vars defined in later imports)
    //---------------------------------------------------------------------
    init() {

        if (GAME_MODE == 'webserver') { this._fullScreen = false; return }
        if (GAME_MODE == 'winjs') {
            this._fullScreen = window.parent.view.isFullScreenMode;
            this._pollController = 0;
            return;
        } else if (GAME_MODE == "webkit") {
            this._fullScreen = true; return;
        } else {
            this._fullScreen = false; return;
        }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::MOUSEMOVE ]----------------------------------
    /*
        PURPOSE: Called in response to user moving the mouse inside the
        application window, populates both raw mouse coords and
        adjusted 'gameWindow' mouse coords, as well as mouseActive flag
    */
    //---------------------------------------------------------------------
    mouseMove(e) {

        this._lastGameMouseX = this._gameMouseX;
        this._lastGameMouseY = this._gameMouseY;
      
        if (e) {
            this._rawMouseX = e.clientX;
            this._rawMouseY = e.clientY;

            this._screenMouseX = e.screenX;
            this._screenMouseY = e.screenY;
        }

        this._gameMouseX = this._rawMouseX - GAME_WINDOW.GAME_WINDOW_LEFT;

        if (this._gameMouseX < 0 ||
            this._gameMouseX > (GAME_WINDOW.GAME_WINDOW_WIDTH)) {
            //sets mouseActive to false if outside the game window
            this._mouseActive = 0;
        } else {
            this._mouseActive = 1;
        }
        this._gameMouseY = this._rawMouseY - GAME_WINDOW.GAME_WINDOW_TOP;

        if (this._gameMouseY < 0 ||
            this._gameMouseY > GAME_WINDOW.GAME_WINDOW_HEIGHT) {
            //sets mouseActive to false if outside the game window
            this._mouseActive = 0;
        }
        
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::MOUSEENTER ]---------------------------------
    /*
        PURPOSE: Called in response to user returning mouse to applicaiton
        window.  Sets mouseActive flag.
    */
    //---------------------------------------------------------------------
    mouseEnter(e) {
        this._mouseActive = 1;
    }

    //-------[ INPUT_MANAGER::MOUSELEAVE ]---------------------------------
    /*
        PURPOSE: Called in response to user moving mouse out of 
        applicaiton window.  Sets mouseActive flag.
    */
    //---------------------------------------------------------------------
    mouseLeave(e) {
        this._mouseActive = 0;
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::MOUSECLICK ]---------------------------------
    /*
        Callback when the user clicks the mouse, we check to see if 
        we are on a menu state.  If we are, then if we clicked outside 
        of any button or control, then keep focus on the current element 
        (i.e. pass it back);
    */
    //---------------------------------------------------------------------
    mouseClick(e) {
        //TODO: Mouse Click functionality
        if (DEBUG && DEBUG_DISPLAY) { DEBUG_DISPLAY.setMouseEvent(); }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::MOUSEDOWN ]---------------------------------
    /*
          Callback when the mouseDown event is fired;
    */
    //---------------------------------------------------------------------
    mouseDown(e) {
        if (e.button == 0) {
            this._mouseState[0] = 1;
        }
        if (e.button == 1) {
            this._mouseState[1] = 1;
        }
        if (e.button == 2) {
            this._mouseState[2] = 1;
        }
        if (DEBUG && DEBUG_DISPLAY) { DEBUG_DISPLAY.setMouseEvent(); }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::MOUSEUP ]---------------------------------
    /*
          Callback when the mouseUp event is fired;
    */
    //---------------------------------------------------------------------
    mouseUp(e) {
        //TODO: Mouse Down functionality
        if (e.button == 0) {
            this._mouseState[0] = 0;
        }
        if (e.button == 1) {
            this._mouseState[1] = 0;
        }
        if (e.button == 2) {
            this._mouseState[2] = 0;
        }
        if (DEBUG && DEBUG_DISPLAY) { DEBUG_DISPLAY.setMouseEvent(); }
    }
    //---------------------------------------------------------------------

    //-------[ INPUT_MANAGER::PROCESSKEYS ]--------------------------------
    /*
        PURPOSE: Bound to keyPress events.
    */
    //---------------------------------------------------------------------
    processKeyPress(e) {
        //console.log("KEYPRESS : " + e.char + " " + e.charCode + " " + e);

        // F is for FULLSCREEN
        if (e.char == 'f' || e.char == 'F' || e.charCode == 102) {
            if (this._text_controls[0] == false) {
                //F key -- Fullscreen Toggle
                if (GAME_MODE == 'webserver') {
                    if (!this._fullScreen) {
                        this.launchIntoFullscreen(document.documentElement);
                        this._fullScreen = true;
                    } else {
                        this.exitFullscreen();
                        this._fullScreen = false;
                    }
                } else if (GAME_MODE == 'winjs') {
                    if (!this._fullScreen) {
                        window.parent.view.tryEnterFullScreenMode();
                        this._fullScreen = true;
                    } else {
                        window.parent.view.exitFullScreenMode();
                        this._fullScreen = false;
                    }
                } else {
                    //TODO: NODE WEBKIT FULLSCREEN TOGGLE...
                }

                e.preventDefault();
                this._text_controls[0] = true; //mark as pressed
            }
        }
        //Q IS FOR DEBUG
        if (e.char == 'q' || e.char == 'Q' || e.charCode == 113) {
            if (this._text_controls[1] == false) {
                if (DEBUG) { DEBUG_DISPLAY.toggleDebugDisplay(); }
                this._text_controls[1] = true;
            }
        }
        //C IS FOR COLLISION DEBUG SCREEN
        if (e.char == 'c' || e.char == 'C' || e.charCode == 99) {
            if (this._text_controls[2] == false) {
                if (DEBUG) { DEBUG_DISPLAY.toggleDebugCanvas(); }
                this._text_controls[2] = true;
            }
        }

        //P IS FOR PAUSE
        if (e.char == 'p' || e.char == 'P' || e.charCode == 112) {
            if (this._text_controls[3] == false) {
                //TODO: IMPLEMENT A CALL TO A PAUSE ROUTING
                this._text_controls[3] = true;
            }
        }

        //ENTER IS FOR SELECTION
        if (e.charCode == 13) {
            if (this._text_controls[4] == false) {
                //TODO: IMPLEMENT SELECTION OF CURRENT ITEM
                this._text_controls[4] = true;
            }
        }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::PROCESSKEYS ]--------------------------------
    /*
        PURPOSE: Bound to keyUp and keyDown events.
    */
    //---------------------------------------------------------------------
    processKeys(e) {
        if (e.keyCode < 200) { //trap against controllers and strange input devices

            //use keyUp and keyDown to track states on arrowkeys, W/A/S/D and space
            if (e.type == 'keydown') {

                if (e.keyCode == 38 || e.keyCode == 87) { this._key_status[0] = 1; }
                if (e.keyCode == 40 || e.keyCode == 83) { this._key_status[1] = 1; }
                if (e.keyCode == 39 || e.keyCode == 68) { this._key_status[2] = 1; }
                if (e.keyCode == 37 || e.keyCode == 65) { this._key_status[3] = 1; }
                if (e.keyCode == 32) { this._key_status[4] = 1; e.preventDefault(); }
                if (e.keyCode == 9) { e.preventDefault(); } //block tab keydown actions

                if (e.keyCode == 40) {
                    e.preventDefault(); //block 'down arrow' for Firefox
                }
            }
            if (e.type == 'keyup') {

                if (e.keyCode == 38 || e.keyCode == 87) { this._key_status[0] = 0; }
                if (e.keyCode == 40 || e.keyCode == 83) { this._key_status[1] = 0; }
                if (e.keyCode == 39 || e.keyCode == 68) { this._key_status[2] = 0; }
                if (e.keyCode == 37 || e.keyCode == 65) { this._key_status[3] = 0; }
                if (e.keyCode == 32) { this._key_status[4] = 0; e.preventDefault(); }
                //TODO: ADJUST SPACE IF ALSO USED AS SELECTION IN MENUS

                //other keys of interest for key up (ie not press-and-hold-style)
                if (e.keyCode == 81) {  //'Q' KEY IS SECRET DEBUG MENU
                    this._text_controls[1] = false;
                }
                if (e.keyCode == 67) {
                    //"PRESSED C" KEY - COLLISION DETECTION DEBUG SCREEN
                    this._text_controls[2] = false;
                    e.preventDefault();
                }
                //pause the game - 'P' key
                if (e.keyCode == 80) {
                    this._text_controls[3] = false;
                    e.preventDefault();
                }
                if (e.keyCode == 13) {
                    //ENTER key -- special case selection trigger?
                    this._text_controls[4] = false;
                    e.preventDefault();
                }
                if (e.keyCode == 9) {
                    //TAB key -- advance menu selection ?
                    //NOTE: Have to only bind to key-up as TAB 
                    //does not always throw a keyPressed event

                    //TODO: ANY POSSIBLE ACTION ON TAB / MENU
                    e.preventDefault();
                }
                //F KEY for fullscreen toggle
                if (e.keyCode == 70) {
                    this._text_controls[0] = false;
                    e.preventDefault();
                }

                //handle escape key for returning to main menu
                if (e.keyCode == 27) {
                    //TODO: GO TO MAIN MENU
                    //TODO: REMOVE QUIT HERE
                    e.preventDefault();
                    main_exit();
                }
            }

            //if we are debugging, tell the debug display
            if (DEBUG && DEBUG_DISPLAY) { DEBUG_DISPLAY.setKeyboardEvent(); }

        } else {
            //nothing.  This trap is here because on newer versions of
            //the winJS and Edge frameworks, events from 360 controllers
            //throw key events with keyCodes in the low 200 range.  (?)
        }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::CONNECTCONTROLLER ]--------------------------
    /*
        Called when a controller is connected or at startup, detects all 
        controllers that can be found through navigator.getGamepads and
        binds up to 2 to the gamepad1 and gamepad2 designations.
    */
    //---------------------------------------------------------------------
    connectController(e) {
        
        if (navigator.getGamepads) {
            for (var i = 0; i < 24; i++) {
                this._gamepads[i] = navigator.getGamepads()[i];
                //this._gamepads[i] = navigator.getGamepads()[i];
                if (!this._gamepads[i]) {
                    this._gamepads[i] = 0;
                } else {
                    //detected controller, so bind it to a player controller...

                    //if player1 doesn't exist, bind controller to player 1
                    if (!this._gamepad_one) {
                        this._gamepad_one = this._gamepads[i];
                    }

                    //if player1 exists but player two doesnt, bind to player 2
                    if (this._gamepad_one && !this.gamepad_two) {
                        if (this._gamepad_one.index != this._gamepads[i].index) {
                            this._gamepad_two = this._gamepads[i];
                        }
                    }
                }
            }
        }
    }
    //---------------------------------------------------------------------
    

    //-------[ INPUT_MANAGER::DISCONNECTCONTROLLER ]-----------------------
    /*
        Called when a controller is disconnected, removes the controller
        and any appropriate binding to player 1 or 2.
    */
    //---------------------------------------------------------------------
    disconnectController(e) {
        console.log("AAENGINE: CONTROLLER DISCONNECTED")
        //if this controller is bound to player 1, remove binding
        var j = e.gamepad.index;

        if (this._gamepad_one) {
            if (this._gamepad_one.index == j) {
                this._gamepad_one = 0;
            }
        }

        //if this controller is bound to player 2, remove binding
        if (this._gamepad_two) {
            if (this._gamepad_two.index == j) {
                this._gamepad_two = 0;
            }
        }

        this._gamepads[j] = 0; //delete from list of gamepads
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER::MOUSEWHEEL ]---------------------------------
    /*
        Update the input manager per-frame.  This function is called
        per-frame from the application loop, and reads in gamepad data,
        and dispatches events or data to other parts of the application 
        as needed.
    */
    //---------------------------------------------------------------------
    mouseWheel(e, isMoz) {
        //get the mouse wheel delta
        var delta = Math.max(-1, Math.min(1, (e.wheelDelta || -e.detail)));

        //add to pool storage (as a way of de-sensitizing mosue a bit)
        this._wheelDeltaStore += delta;
        this._totalWheelDeltaStore += delta;

        if (Math.abs(this._wheelDeltaStore) > this._mouseWheelSensitivity) {

            if (this._wheelDeltaStore < 0) {
                this._mouseState[3] = -1;
            } else {
                this._mouseState[3] = 1;
            }        
            this._wheelDeltaStore = 0;
        } 
        e.preventDefault(); //prevent the mouse wheel from normal action
        if (DEBUG && DEBUG_DISPLAY) { DEBUG_DISPLAY.setMouseWheelEvent(true); }
    }

    //-------[ INPUT_MANAGER::UPDATE ]-------------------------------------
    /*
        Update the input manager per-frame.  This function is called
        per-frame from the application loop, and reads in gamepad data,
        and dispatches events or data to other parts of the application 
        as needed.
    */
    //---------------------------------------------------------------------
    update() {

        if (GAME_MODE == "winjs" && this._pollController == 60) {
            this.connectController();
            this._pollController = 0;

        } else {
            this._pollController++;
        }

        //clear the wheel delta if we returned a positive event last frame
        if (this._wheelClearDeltaStore) {
            this._wheelDeltaStore = 0;
            this._mouseState[3] = 0;
            this._wheelClearDeltaStore = 0;
        }

        //-----------------------------------------------------------------
        // PROCESS GAMEPAD ONE
        //-----------------------------------------------------------------
        if (this._gamepad_one) {
            //get new vales from the controller
            this._gamepad_one = navigator.getGamepads()[
                this._gamepad_one.index];
        }

        if (this._gamepad_one) {

            //connection
            this._gamepad_one_data[0] = this._gamepad_one.connected;

            //mapping
            if (this._gamepad_one.mapping == 'standard') {
                this._gamepad_one_data[1] = 1;
            } else { this._gamepad_one_data[1] = 0; }

            //id
            this._gamepad_one_data[2] = this._gamepad_one.id;

            //axis - and deadzone processing
            this._gamepad_one_data[3] = this._gamepad_one.axes[0];
            this._gamepad_one_data[4] = this._gamepad_one.axes[1];
            this._gamepad_one_data[5] = this._gamepad_one.axes[2];
            this._gamepad_one_data[6] = this._gamepad_one.axes[3];

            if (Math.abs(this._gamepad_one_data[3]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_one_data[3] = 0;
            }
            if (Math.abs(this._gamepad_one_data[4]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_one_data[4] = 0;
            }
            if (Math.abs(this._gamepad_one_data[5]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_one_data[5] = 0;
            }
            if (Math.abs(this._gamepad_one_data[6]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_one_data[6] = 0;
            }

            //read and process left and right trigger
            this._gamepad_one_data[13] = this._gamepad_one.buttons[6].value;
            this._gamepad_one_data[14] = this._gamepad_one.buttons[7].value;
            if (Math.abs(this._gamepad_one_data[13]) < TRIGGER_DEAD_ZONE) {
                this._gamepad_one_data[13] = 0;
            }
            if (Math.abs(this._gamepad_one_data[14]) < TRIGGER_DEAD_ZONE) {
                this._gamepad_one_data[14] = 0;
            }

            //A, B, X, Y buttons (in order)
            this._gamepad_one_data[7]  = this._gamepad_one.buttons[0].value;
            this._gamepad_one_data[8]  = this._gamepad_one.buttons[1].value;
            this._gamepad_one_data[9]  = this._gamepad_one.buttons[2].value;
            this._gamepad_one_data[10] = this._gamepad_one.buttons[3].value;

            //left and right bumper
            this._gamepad_one_data[11] = this._gamepad_one.buttons[4].value;
            this._gamepad_one_data[12] = this._gamepad_one.buttons[5].value;
            
            //all the other buttons and d-pad
            this._gamepad_one_data[15] = this._gamepad_one.buttons[8].value;
            this._gamepad_one_data[16] = this._gamepad_one.buttons[9].value;
            this._gamepad_one_data[17] = this._gamepad_one.buttons[10].value;
            this._gamepad_one_data[18] = this._gamepad_one.buttons[11].value;
            this._gamepad_one_data[19] = this._gamepad_one.buttons[12].value;
            this._gamepad_one_data[20] = this._gamepad_one.buttons[13].value;
            this._gamepad_one_data[21] = this._gamepad_one.buttons[14].value;
            this._gamepad_one_data[22] = this._gamepad_one.buttons[15].value;

        } else {
            for (var i = 0; i < this._gamepad_one_data.length; i++) {
                this._gamepad_one_data[i] = 0;
            }
        }

        //-----------------------------------------------------------------
        // PROCESS GAMEPAD TWO
        //-----------------------------------------------------------------
        if (this._gamepad_two) {
            //get new vales from the controller
            this._gamepad_two = navigator.getGamepads()[
                this._gamepad_two.index];
        }

        if (this._gamepad_two) {

            //connection
            this._gamepad_two_data[0] = this._gamepad_two.connected;

            //mapping
            if (this._gamepad_two.mapping == 'standard') {
                this._gamepad_two_data[1] = 1;
            } else { this._gamepad_two_data[1] = 0; }

            //id
            this._gamepad_two_data[2] = this._gamepad_two.id;

            //axis - and deadzone processing
            this._gamepad_two_data[3] = this._gamepad_two.axes[0];
            this._gamepad_two_data[4] = this._gamepad_two.axes[1];
            this._gamepad_two_data[5] = this._gamepad_two.axes[2];
            this._gamepad_two_data[6] = this._gamepad_two.axes[3];

            if (Math.abs(this._gamepad_two_data[3]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_two_data[3] = 0;
            }
            if (Math.abs(this._gamepad_two_data[4]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_two_data[4] = 0;
            }
            if (Math.abs(this._gamepad_two_data[5]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_two_data[5] = 0;
            }
            if (Math.abs(this._gamepad_two_data[6]) < CONTROLLER_DEAD_ZONE) {
                this._gamepad_two_data[6] = 0;
            }

            //read and process left and right trigger
            this._gamepad_two_data[13] = this._gamepad_two.buttons[6].value;
            this._gamepad_two_data[14] = this._gamepad_two.buttons[7].value;
            if (Math.abs(this._gamepad_two_data[13]) < TRIGGER_DEAD_ZONE) {
                this._gamepad_two_data[13] = 0;
            }
            if (Math.abs(this._gamepad_two_data[14]) < TRIGGER_DEAD_ZONE) {
                this._gamepad_two_data[14] = 0;
            }

            //A, B, X, Y buttons (in order)
            this._gamepad_two_data[7]  = this._gamepad_two.buttons[0].value;
            this._gamepad_two_data[8]  = this._gamepad_two.buttons[1].value;
            this._gamepad_two_data[9]  = this._gamepad_two.buttons[2].value;
            this._gamepad_two_data[10] = this._gamepad_two.buttons[3].value;

            //left and right bumper
            this._gamepad_two_data[11] = this._gamepad_two.buttons[4].value;
            this._gamepad_two_data[12] = this._gamepad_two.buttons[5].value;

            //all the other buttons and d-pad
            this._gamepad_two_data[15] = this._gamepad_two.buttons[8].value;
            this._gamepad_two_data[16] = this._gamepad_two.buttons[9].value;
            this._gamepad_two_data[17] = this._gamepad_two.buttons[10].value;
            this._gamepad_two_data[18] = this._gamepad_two.buttons[11].value;
            this._gamepad_two_data[19] = this._gamepad_two.buttons[12].value;
            this._gamepad_two_data[20] = this._gamepad_two.buttons[13].value;
            this._gamepad_two_data[21] = this._gamepad_two.buttons[14].value;
            this._gamepad_two_data[22] = this._gamepad_two.buttons[15].value;

        } else {
            for (var i = 0; i < this._gamepad_two_data.length; i++) {
                this._gamepad_two_data[i] = 0;
            }
        }

        //NOTE: Only basic key_status, simple mouse and gamepad data 
        //is currently surfaced to game_objects / the threaded environment, 
        //as menus are likely in the main UI thread.  this can be changed 
        //later if necessary for feature implementaiton
        if (USE_THREADING) {

            this._mouse_data[0] = this._mouseActive;
            this._mouse_data[1] = this._gameMouseX;
            this._mouse_data[2] = this._gameMouseY;

            THREAD_MANAGER.sendInput([this._key_status, this._gamepad_one_data,
            this._gamepad_two_data, this._mouse_data]);
        }
    }
    //---------------------------------------------------------------------

    //-------[ INPUT_MANAGER : ENTER FULL SCREEN ]-------------------------
    // enter full screen mode on a given element (generally the document)
    //---------------------------------------------------------------------
    launchIntoFullscreen(element) {
        if (element.requestFullscreen) {
            element.requestFullscreen();
           
        } else if (element.mozRequestFullScreen) {
            element.mozRequestFullScreen();
           
        } else if (element.webkitRequestFullscreen) {
            element.webkitRequestFullscreen();
            
        } else if (element.msRequestFullscreen) {
            element.msRequestFullscreen();
            
        }
    }
    //---------------------------------------------------------------------

    //-------[ INPUT_MANAGER : EXIT FULL SCREEN ]--------------------------
    // called to exit full screen mode
    //---------------------------------------------------------------------
    exitFullscreen() {
        if (document.exitFullscreen) {
            document.exitFullscreen();
        } else if (document.mozCancelFullScreen) {
            document.mozCancelFullScreen();
        } else if (document.webkitExitFullscreen) {
            document.webkitExitFullscreen();
        }
    }
    //---------------------------------------------------------------------


    //-------[ INPUT_MANAGER : resize ]------------------------------------
    // called in response to resize event, handles adjusting mouse coords
    // as window is no longer capable of tracking adjustment 
    //---------------------------------------------------------------------
    resize(e) {
        if (GAME_STATE >= 4) {

            //this works in mozilla and chrome, not winJS
            //no idea why 88 is a number here, title/tab bar?!?
            if (GAME_MODE != "winjs") {
                this._gameMouseX = Math.round(
                    this._screenMouseX - window.screenX);
                if (this._fullScreen) {
                    this._rawMouseY = Math.round(this._screenMouseY -
                    window.screenY);
                    this._gameMouseY = this._rawMouseY
                        - GAME_WINDOW.GAME_WINDOW_TOP;
                } else {
                    this._rawMouseY = Math.round(this._screenMouseY -
                    window.screenY - 88);
                    this._gameMouseY = this._rawMouseY
                        - GAME_WINDOW.GAME_WINDOW_TOP;
                }

            } else {
                //winjs / edge compatible way of doing the same thing
                var sz = (screen.deviceXDPI / screen.logicalXDPI);
                this._rawMouseX = Math.round((this._screenMouseX / sz)
                    - window.screenLeft * sz);
                this._gameMouseX = this._rawMouseX;

                this._rawMouseY = Math.round((this._screenMouseY / sz)
                    - window.screenTop * sz);
                this._gameMouseY = this._rawMouseY
                    - GAME_WINDOW.GAME_WINDOW_TOP;
            }
        }
    }

    //-------[ INPUT_MANAGER ACCESS METHODS ]------------------------------
    //mouse
    get MOUSE_X() { return this._gameMouseX; }
    get MOUSE_Y() { return this._gameMouseY; }
    get LAST_MOUSE_X() { return this._lastGameMouseX; }
    get LAST_MOUSE_Y() { return this._lastGameMouseY; }
    get RAW_MOUSE_X() { return this._rawMouseX; }
    get RAW_MOUSE_Y() { return this._rawMouseY; }
    get MOUSE_ACTIVE() { return this._mouseActive; }
    get MOUSE_LEFT() { return this._mouseState[0]; }
    get MOUSE_MIDDLE() { return this._mouseState[1]; }
    get MOUSE_RIGHT() { return this._mouseState[2]; }
    get MOUSE_WHEEL_TOTAL_DELTA() { return this._totalWheelDeltaStore; }
    get MOUSE_WHEEL() {
        var mwe = this._mouseState[3];
        if (mwe < 0) {  this._wheelClearDeltaStore = 1; return -1; }
        if (mwe > 0) {  this._wheelClearDeltaStore = 1; return 1; }
        return 0;
    }

    //keyboard
    get KB_UP() { return this._key_status[0]; }
    get KB_DOWN() { return this._key_status[1]; }
    get KB_LEFT() { return this._key_status[3]; }
    get KB_RIGHT() { return this._key_status[2]; }
    get KB_SPACE() { return this._key_status[4]; }

    //gamepad 1-2  
    //to get the entire state of the controller (for functions that
    //use several elements in conjunction), just get the entire data
    //array for the frame. 
    get GAMEPAD1() { return this._gamepad_one_data; }
    get GAMEPAD2() { return this._gamepad_two_data; }

    //or use individual handlers for a given field
    get GAMEPAD1_CON()   { return this._gamepad_one_data[0]; }
    get GAMEPAD1_MAP()   { return this._gamepad_one_data[1]; }
    get GAMEPAD1_ID()    { return this._gamepad_one_data[2]; }
    get GAMEPAD1_LHZAXIS() { return this._gamepad_one_data[3]; }
    get GAMEPAD1_LVTAXIS() { return this._gamepad_one_data[4]; }
    get GAMEPAD1_RHZAXIS() { return this._gamepad_one_data[5]; }
    get GAMEPAD1_RVTAXIS() { return this._gamepad_one_data[6]; }
    get GAMEPAD1_LTRIGGER() { return this._gamepad_one_data[13]; }
    get GAMEPAD1_RTRIGGER() { return this._gamepad_one_data[14]; }
    get GAMEPAD1_LBUMPER() { return this._gamepad_one_data[11]; }
    get GAMEPAD1_RBUMPER() { return this._gamepad_one_data[12]; }
    get GAMEPAD1_A() { return this._gamepad_one_data[7]; }
    get GAMEPAD1_B() { return this._gamepad_one_data[8]; }
    get GAMEPAD1_X() { return this._gamepad_one_data[9]; }
    get GAMEPAD1_Y() { return this._gamepad_one_data[10]; }
    get GAMEPAD1_DUP() { return this._gamepad_one_data[19]; }
    get GAMEPAD1_DDOWN() { return this._gamepad_one_data[20]; }
    get GAMEPAD1_DRIGHT() { return this._gamepad_one_data[21]; }
    get GAMEPAD1_DLEFT() { return this._gamepad_one_data[22]; }
    get GAMEPAD1_BACK() { return this._gamepad_one_data[15]; }
    get GAMEPAD1_START() { return this._gamepad_one_data[16]; }
    get GAMEPAD1_LSTICK() { return this._gamepad_one_data[17]; }
    get GAMEPAD1_RSTICK() { return this._gamepad_one_data[18]; }

    get GAMEPAD2_CON() { return this._gamepad_two_data[0]; }
    get GAMEPAD2_MAP() { return this._gamepad_two_data[1]; }
    get GAMEPAD2_ID() { return this._gamepad_two_data[2]; }
    get GAMEPAD2_LHZAXIS() { return this._gamepad_two_data[3]; }
    get GAMEPAD2_LVTAXIS() { return this._gamepad_two_data[4]; }
    get GAMEPAD2_RHZAXIS() { return this._gamepad_two_data[5]; }
    get GAMEPAD2_RVTAXIS() { return this._gamepad_two_data[6]; }
    get GAMEPAD2_LTRIGGER() { return this._gamepad_two_data[13]; }
    get GAMEPAD2_RTRIGGER() { return this._gamepad_two_data[14]; }
    get GAMEPAD2_LBUMPER() { return this._gamepad_two_data[11]; }
    get GAMEPAD2_RBUMPER() { return this._gamepad_two_data[12]; }
    get GAMEPAD2_A() { return this._gamepad_two_data[7]; }
    get GAMEPAD2_B() { return this._gamepad_two_data[8]; }
    get GAMEPAD2_X() { return this._gamepad_two_data[9]; }
    get GAMEPAD2_Y() { return this._gamepad_two_data[10]; }
    get GAMEPAD2_DUP() { return this._gamepad_two_data[19]; }
    get GAMEPAD2_DDOWN() { return this._gamepad_two_data[20]; }
    get GAMEPAD2_DRIGHT() { return this._gamepad_two_data[21]; }
    get GAMEPAD2_DLEFT() { return this._gamepad_two_data[22]; }
    get GAMEPAD2_BACK() { return this._gamepad_two_data[15]; }
    get GAMEPAD2_START() { return this._gamepad_two_data[16]; }
    get GAMEPAD2_LSTICK() { return this._gamepad_two_data[17]; }
    get GAMEPAD2_RSTICK() { return this._gamepad_two_data[18]; }
  
}
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// global handle to input manager, used throughout the application
var INPUT_MANAGER = INPUT_MANAGER || new InputManager();
//-------------------------------------------------------------------------

//--------[ EVENT LISTENER BINDINGS ]--------------------------------------
//-------------------------------------------------------------------------
window.addEventListener('mousemove', function (e)
                         { INPUT_MANAGER.mouseMove(e) },            false);
//-------------------------------------------------------------------------
window.addEventListener('mouseover', function (e)
                         { INPUT_MANAGER.mouseEnter(e) },           false);
//-------------------------------------------------------------------------
window.addEventListener('mouseout', function (e)
                         { INPUT_MANAGER.mouseLeave(e) },           false);
//-------------------------------------------------------------------------
window.addEventListener('mouseup', function (e)
                         { INPUT_MANAGER.mouseUp(e) },              false);
//-------------------------------------------------------------------------
window.addEventListener('mousedown', function (e)
                         { INPUT_MANAGER.mouseDown(e) },            false);
//-------------------------------------------------------------------------
window.addEventListener('mouseclick', function (e)
                         { INPUT_MANAGER.mouseClick(e) },           false);
//-------------------------------------------------------------------------
window.addEventListener('keyup', function (e)
                         { INPUT_MANAGER.processKeys(e) },          false);
//-------------------------------------------------------------------------
window.addEventListener('keydown', function (e)
                         { INPUT_MANAGER.processKeys(e) },          false);
//-------------------------------------------------------------------------
window.addEventListener('keypress', function (e)
                         { INPUT_MANAGER.processKeyPress(e) },      false);
//-------------------------------------------------------------------------
window.addEventListener("gamepaddisconnected", function (e)
                         { INPUT_MANAGER.disconnectController(e) }, false);
//-------------------------------------------------------------------------
window.addEventListener("gamepadconnected", function (e)
                         { INPUT_MANAGER.connectController(e) },    false);
//-------------------------------------------------------------------------
window.addEventListener("mousewheel", function (e)
                         { INPUT_MANAGER.mouseWheel(e, true) },     false);
//-------------------------------------------------------------------------
window.addEventListener("DOMMouseScroll", function (e)
                         { INPUT_MANAGER.mouseWheel(e, false) },    false);
//-------------------------------------------------------------------------


