﻿//-----[ THREADED GAME OBJECT MANAGER ]------------------------------------
/* PURPOSE: The threaded game object manager re-creates some of the basic
environment for game objects, but in a worker thread.  This moves a bunch
of the math and calculation for object updates and collision detection
OFF of the UI thread.

The core here is that there is a threaded_game_objects[] array to hold the
game objects, and a threaded_objects_min_data[] array that is the minimual
information set that needs to be sent back to the main thread to draw and
render the objects in the 3D world.
*/
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
var THREAD_TIMING = 16.6667;  //goal time to update thread on
var SCALED_THREAD_TIMING = 0; //scaled time
var THREAD_DEBUG_SCALE = 0.1; //modifier for debug mode
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
//local copies of GAME_WINDOW and associated vars for this thread pool
//these are updated from the real GAME_WINDOW on a resize event
//-------------------------------------------------------------------------
var GAME_WINDOW_ZOOM_EFFECT = 0;
var GAME_WINDOW = new Object();
var INPUT_MANAGER = new Object();
var AACOLLISION_MANAGER = 0; //pointer for redirection...
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
//local copies of other needed vars
//-------------------------------------------------------------------------
var DEBUG = true;
var DEBUG_CANVAS_VISIBILITY = false;
var USE_PERFORMANCE_TIMING = false; //set to true if capability exists
var USE_THREADING = true; //we're in a thread, so this is a safe assumption
var sinTable = [];
var cosTable = [];
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
//array to hold information on input from the main thread
//-------------------------------------------------------------------------
var a = new Float32Array(4);
var b = c = new Float32Array(23);
var d = new Int32Array(3);
var INPUT_INFO = [a,b,c,d];
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
//import common scripts needed for threaded support
//-------------------------------------------------------------------------
importScripts('../../libraries/three/three.js');
importScripts('../GameObject.js');
importScripts('../CollisionManager.js');
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
//import specific game objects so they can be used in thread land
//-------------------------------------------------------------------------
importScripts('../../game_objects/Bouncer.js');
importScripts('../../game_objects/MouseFollower.js');
importScripts('../../game_objects/InputFollower.js');
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
function threadFrame() {
    
    if (this._active) {
        //-----------------------------------------------------------------
        // store time at the start of this threadFrame
        //-----------------------------------------------------------------
        SCALED_THREAD_TIMING = THREAD_TIMING;

        if (USE_PERFORMANCE_TIMING) {
            this._thread_date = performance.now();
        } else {
            this._thread_date = Date.now();
        }

        //-----------------------------------------------------------------
        //call PRE-UPDATE for collision manager (clear bins)
        //-----------------------------------------------------------------
        AACOLLISION_MANAGER.preupdate();


        //-----------------------------------------------------------------
        //for each threaded object, update, render, and pack information
        //into a transfer array
        //-----------------------------------------------------------------
        for (var i = 0; i < this._threaded_objects.length; i++) {

            //update this object
            this._threaded_objects[i].update();
            //render this object
            this._threaded_objects[i].render();
            
            //pack updated information into min array for transfer
            this._threaded_objects_min_data[i][0][0] =
                this._threaded_objects[i]._position.x;
            this._threaded_objects_min_data[i][0][1] =
               this._threaded_objects[i]._position.y;
            this._threaded_objects_min_data[i][0][2] =
               this._threaded_objects[i]._position.z;

            this._threaded_objects_min_data[i][1][0] =
                this._threaded_objects[i]._rotation.x;
            this._threaded_objects_min_data[i][1][1] =
                this._threaded_objects[i]._rotation.y;
            this._threaded_objects_min_data[i][1][2] =
                this._threaded_objects[i]._rotation.z;

            this._threaded_objects_min_data[i][2][0] =
                this._threaded_objects[i]._meshScale.x;
            this._threaded_objects_min_data[i][2][1] =
                this._threaded_objects[i]._meshScale.y;
            this._threaded_objects_min_data[i][2][2] =
                this._threaded_objects[i]._meshScale.z;
                

            this._threaded_objects_min_data[i][3] = this._threaded_objects[i]._meshID;
            this._threaded_objects_min_data[i][4] = 99; //nonexistent type
            //this._threaded_objects_min_data[i][4] = 0;
            //this._threaded_objects_min_data[i][5].length = 0; //clear
            //this._threaded_objects_min_data[i][6] = false;

            //-----------------------------------------------------------------
            //if debugging and debug canvas, prep and pack to draw 2D overlays
            //-----------------------------------------------------------------
            if (DEBUG) {
                if (DEBUG_CANVAS_VISIBILITY) {
                    //store the type of the object for parsing
                    this._threaded_objects_min_data[i][4] =
                        this._threaded_objects[i]._type;

                    //if a poly game object
                    if (this._threaded_objects[i]._type == 2) {
                        this._threaded_objects_min_data[i][5][0] =
                            this._threaded_objects[i]._projected2dPosX; //x
                        this._threaded_objects_min_data[i][5][1] =
                            this._threaded_objects[i]._projected2dPosY, //y
                        this._threaded_objects_min_data[i][5][2] =
                            this._threaded_objects[i]._adjusted_points; //polypoints
                    }

                    //if a circle game object
                    if (this._threaded_objects[i]._type == 1) {
                        this._threaded_objects_min_data[i][5][0] =
                            this._threaded_objects[i]._projected2dPosX; //x
                        this._threaded_objects_min_data[i][5][1] =
                            this._threaded_objects[i]._projected2dPosY; //y
                        this._threaded_objects_min_data[i][5][2] =
                            this._threaded_objects[i]._radius;
                    }

                    //if a box game object
                    if (this._threaded_objects[i]._type == 0) {
                        this._threaded_objects_min_data[i][5][0] =
                            this._threaded_objects[i]._projected2dPosX -
                            this._threaded_objects[i]._hw *
                            GAME_WINDOW.GAME_WINDOW_SCALE, //x
                        this._threaded_objects_min_data[i][5][1] =
                            this._threaded_objects[i]._projected2dPosY -
                            this._threaded_objects[i]._hh *
                            GAME_WINDOW.GAME_WINDOW_SCALE; //y
                        this._threaded_objects_min_data[i][5][2] =
                            this._threaded_objects[i]._width *
                            GAME_WINDOW.GAME_WINDOW_SCALE;
                        this._threaded_objects_min_data[i][5][3] =
                            this._threaded_objects[i]._height *
                            GAME_WINDOW.GAME_WINDOW_SCALE;
                    }
                }
            }
        }

        //-----------------------------------------------------------------
        //calc out how long all of this took, and then tack it on to our data
        //post game object update, pre-collision call
        //-----------------------------------------------------------------
        if (DEBUG) {
            if (DEBUG_CANVAS_VISIBILITY) {
                this._threaded_objects_min_data[
                    this._threaded_objects.length] =
                    AACOLLISION_MANAGER._collisionBins;
            } else {
                this._threaded_objects_min_data[
                    this._threaded_objects.length] = 0;
            }
            if (USE_PERFORMANCE_TIMING) {
                this._dt = performance.now() - this._thread_date;
            } else {
                this._dt = Date.now() - this._thread_date;
            }
            this._threaded_objects_min_data[
                this._threaded_objects.length + 1] = this._dt;
        } else {
            this._threaded_objects_min_data[
                this._threaded_objects.length + 1] = 0;
        }

        //-----------------------------------------------------------------
        //update collision manager
        //-----------------------------------------------------------------
        AACOLLISION_MANAGER.update();

        //-----------------------------------------------------------------
        //calc out how long all of this took, and then tack it on to our data
        //post collision call...
        //-----------------------------------------------------------------
        if (DEBUG) {

            if (DEBUG_CANVAS_VISIBILITY) {
                SCALED_THREAD_TIMING = THREAD_TIMING * THREAD_DEBUG_SCALE;

                for (var j = 0; j < this._threaded_objects.length; j++) {
                    //console.log(j);
                    this._threaded_objects_min_data[j][6] =
                        this._threaded_objects[j]._collidedThisFrame;
                }
            }

            if (USE_PERFORMANCE_TIMING) {
                this._cdt = performance.now() -
                    this._thread_date - this._dt;
            } else {
                this._cdt = Date.now() -
                    this._thread_date - this._dt;
            }
            this._threaded_objects_min_data[
                this._threaded_objects.length + 2] = this._cdt;
            this._threaded_objects_min_data[
                this._threaded_objects.length + 3] = this._inputProcTime;
        } else {
            this._threaded_objects_min_data[
                this._threaded_objects.length + 2] = 0;
            this._threaded_objects_min_data[
                this._threaded_objects.length + 3] = 0;
        }

        //-----------------------------------------------------------------
        //add command code 0 to the end as this is 'main process'
        //-----------------------------------------------------------------
        this._threaded_objects_min_data[this._threaded_objects.length + 4] = 0;

        //-----------------------------------------------------------------
        // set all this back to main UI thread for rendering
        //-----------------------------------------------------------------
        postMessage(this._threaded_objects_min_data);
    }
    //--------------------------------------------------------------------
    
    
    //------[ onMessage handler ]-----------------------------------------
    /* called from the THREAD_MANAGER, takes input and based on control
    flag does one of several routines for initialization or update
    from the main UI thread.
    */
    //--------------------------------------------------------------------
    this.onmessage = function (e) {
        
        if (e.data[0] == 1) {

            //INITIALIZE THIS THREAD TO MANAGE GAME OBJECTS
            SCALED_THREAD_TIMING = THREAD_TIMING;
            this._threaded_objects = [];
            this._threaded_objects_min_data = [];

            //create localized versions of sin/cos tables for use with threads
            for (var s = 0; s < 360; s++) {
                sinTable[s] = Math.sin(s * Math.PI / 180.0000);
                cosTable[s] = Math.cos(s * Math.PI / 180.0000);
            }

            //store creation time of this thread            
            this._thread_date = new Date().getTime();
            this._inputProcTime = 0;

            //mark this thread as an active manager and start updating
            this._active = true;

            AACOLLISION_MANAGER = new CollisionManager();
            AACOLLISION_MANAGER.update();

            try {
                this._thread_date = performance.now();
                USE_PERFORMANCE_TIMING = true;
            } catch (e) {
                this._thread_date = Date.now();
            }

        } else if (e.data[0] == 2) { 
            //RESIZE EVENT - UPDATE ALL NEEDED INFO ON GAME WINDOW
            GAME_WINDOW_ZOOM_EFFECT = e.data[1];
            GAME_WINDOW.GAME_WINDOW_SCALE = e.data[2];
            GAME_WINDOW.GAME_WINDOW_HALF_WIDTH = e.data[3];
            GAME_WINDOW.GAME_WINDOW_HALF_HEIGHT = e.data[4];
            GAME_WINDOW.GAME_WINDOW_UNSCALED_WIDTH = e.data[5];
            GAME_WINDOW.GAME_WINDOW_UNSCALED_HEIGHT = e.data[6];
            GAME_WINDOW.GAME_WINDOW_UNSCALED_HALF_WIDTH = e.data[7];
            GAME_WINDOW.GAME_WINDOW_UNSCALED_HALF_HEIGHT = e.data[8];

            //tell the collision manager new information as needed
            AACOLLISION_MANAGER.init(e.data[9], e.data[10]);

            //call resize on all game objects
            for (var i = 0; i < this._threaded_objects.length; i++) {
                this._threaded_objects[i].resize();
            }

        } else if (e.data[0] == 3) {

            //TURN OFF EVENT - CALLED ON EXIT
            this._active = false;
            clearTimeout("threadFrame()");

        } else if (e.data[0] == 4) {

            //LOAD A GAME OBJECT
            var go = eval('new ' + e.data[1] + '();');

            var id = this._threaded_objects.push(go);
            go.setID(id-1);
            var p = new Float32Array(3);
            var r = new Float32Array(3);
            var s = new Float32Array(3);
            this._threaded_objects_min_data.push([p, r, s, 0, 0, [], false]);

            //callback to note this object is loaded
            postMessage([1]);

        } else if (e.data[0] == 5) {

            //call to toggle debug canvas visibility
            DEBUG_CANVAS_VISIBILITY = e.data[1];

        } else if (e.data[0] == 6) {
            //INPUT UPDATE
            //get input from main thread and process for local
            //threaded environment to use
            if (DEBUG) {
                if (USE_PERFORMANCE_TIMING) {
                    this._inputPT = performance.now();
                } else {
                    this._inputPT = Date.now();
                }
            }
            //store information about keybaord status
            INPUT_MANAGER.KB_UP             = e.data[1][0][0];
            INPUT_MANAGER.KB_DOWN           = e.data[1][0][1];
            INPUT_MANAGER.KB_LEFT           = e.data[1][0][2];
            INPUT_MANAGER.KB_RIGHT          = e.data[1][0][3];
            INPUT_MANAGER.KB_SPACE          = e.data[1][0][4];

            //store information about first game controller
            INPUT_MANAGER.GAMEPAD1_CON      = e.data[1][1][0];
            INPUT_MANAGER.GAMEPAD1_MAP      = e.data[1][1][1];
            INPUT_MANAGER.GAMEPAD1_ID       = e.data[1][1][2];
            INPUT_MANAGER.GAMEPAD1_LHZAXIS  = e.data[1][1][3];
            INPUT_MANAGER.GAMEPAD1_LVTAXIS  = e.data[1][1][4];
            INPUT_MANAGER.GAMEPAD1_RHZAXIS  = e.data[1][1][5];
            INPUT_MANAGER.GAMEPAD1_RVTAXIS  = e.data[1][1][6];
            INPUT_MANAGER.GAMEPAD1_LTRIGGER = e.data[1][1][13];
            INPUT_MANAGER.GAMEPAD1_RTRIGGER = e.data[1][1][14];
            INPUT_MANAGER.GAMEPAD1_LBUMPER  = e.data[1][1][11];
            INPUT_MANAGER.GAMEPAD1_RBUMPER  = e.data[1][1][12];
            INPUT_MANAGER.GAMEPAD1_A        = e.data[1][1][7];
            INPUT_MANAGER.GAMEPAD1_B        = e.data[1][1][8];
            INPUT_MANAGER.GAMEPAD1_X        = e.data[1][1][9];
            INPUT_MANAGER.GAMEPAD1_Y        = e.data[1][1][10];
            INPUT_MANAGER.GAMEPAD1_DUP      = e.data[1][1][19];
            INPUT_MANAGER.GAMEPAD1_DDOWN    = e.data[1][1][20];
            INPUT_MANAGER.GAMEPAD1_DRIGHT   = e.data[1][1][21];
            INPUT_MANAGER.GAMEPAD1_DLEFT    = e.data[1][1][22];
            INPUT_MANAGER.GAMEPAD1_BACK     = e.data[1][1][15];
            INPUT_MANAGER.GAMEPAD1_START    = e.data[1][1][16];
            INPUT_MANAGER.GAMEPAD1_LSTICK   = e.data[1][1][17];
            INPUT_MANAGER.GAMEPAD1_RSTICK   = e.data[1][1][18];
            
            //store information about second game controller
            INPUT_MANAGER.GAMEPAD2_CON      = e.data[1][2][0];
            INPUT_MANAGER.GAMEPAD2_MAP      = e.data[1][2][1];
            INPUT_MANAGER.GAMEPAD2_ID       = e.data[1][2][2];
            INPUT_MANAGER.GAMEPAD2_LHZAXIS  = e.data[1][2][3];
            INPUT_MANAGER.GAMEPAD2_LVTAXIS  = e.data[1][2][4];
            INPUT_MANAGER.GAMEPAD2_RHZAXIS  = e.data[1][2][5];
            INPUT_MANAGER.GAMEPAD2_RVTAXIS  = e.data[1][2][6];
            INPUT_MANAGER.GAMEPAD2_LTRIGGER = e.data[1][2][13];
            INPUT_MANAGER.GAMEPAD2_RTRIGGER = e.data[1][2][14];
            INPUT_MANAGER.GAMEPAD2_LBUMPER  = e.data[1][2][11];
            INPUT_MANAGER.GAMEPAD2_RBUMPER  = e.data[1][2][12];
            INPUT_MANAGER.GAMEPAD2_A        = e.data[1][2][7];
            INPUT_MANAGER.GAMEPAD2_B        = e.data[1][2][8];
            INPUT_MANAGER.GAMEPAD2_X        = e.data[1][2][9];
            INPUT_MANAGER.GAMEPAD2_Y        = e.data[1][2][10];
            INPUT_MANAGER.GAMEPAD2_DUP      = e.data[1][2][19];
            INPUT_MANAGER.GAMEPAD2_DDOWN    = e.data[1][2][20];
            INPUT_MANAGER.GAMEPAD2_DRIGHT   = e.data[1][2][21];
            INPUT_MANAGER.GAMEPAD2_DLEFT    = e.data[1][2][22];
            INPUT_MANAGER.GAMEPAD2_BACK     = e.data[1][2][15];
            INPUT_MANAGER.GAMEPAD2_START    = e.data[1][2][16];
            INPUT_MANAGER.GAMEPAD2_LSTICK   = e.data[1][2][17];
            INPUT_MANAGER.GAMEPAD2_RSTICK   = e.data[1][2][18];

            INPUT_MANAGER.MOUSE_ACTIVE      = e.data[1][3][0];
            INPUT_MANAGER.MOUSE_X           = e.data[1][3][1];
            INPUT_MANAGER.MOUSE_Y           = e.data[1][3][2];

            //calc out how long all of this took, 
            //and then tack it on to our data
            if (DEBUG) {
                if (USE_PERFORMANCE_TIMING) {
                    this._inputProcTime = performance.now() - this._inputPT;
                } else {
                    this._inputProcTime = Date.now() - this._inputPT;
                }   
            }


        } else if (e.data[0] == 7) {
            //REQUEST FOR OBJECT LINKING DATA
            //call to link object id and mesh id
            var data = [];
            for (var i = 0; i < this._threaded_objects.length; i++) {

                data[i] = [this._threaded_objects[i]._id,
                           this._threaded_objects[i]._meshID]
            }
            data.push(2);
            postMessage(data);

        } else {
            //UNKNOWN CONTROL FLAG, GO INACTIVE
            this._active = false;
        }
    }

    this.reset = function () {
        //self.console.log("called reset");
    }

    setTimeout("threadFrame()", SCALED_THREAD_TIMING - this._cdt - this._dt);
}
//-------------------------------------------------------------------------


//-----[ threadFrame() ]---------------------------------------------------
/* callback to start the loop for this thread (thread does nothing 
until it is marked active through init call from onMessage) */
//-------------------------------------------------------------------------
threadFrame();
//-------------------------------------------------------------------------